from calculator import Calculator
from logicbook import Logic
from test_calculator import TestCalculatorSub, TestCalculatorAdd

add = Logic(
    name="Calculator/Add",
    func=Calculator().add,
    readme="calculator_add.md",
)

add.add_example(
    name="Default", 
    func=Calculator().add,
    args={
        "a": 1,
        "b": 2
    })

add.add_test(
  name="Test Class of Calculator", 
  test_case=TestCalculatorAdd
)

sub = Logic(
    name="Calculator/Sub",
    func=Calculator().sub,
    readme="calculator_sub.md",
)

sub.add_example(
    name="Default", 
    func=Calculator().sub,
    args={
        "a": 2,
        "b": 1
    })

sub.add_test(
  name="Test Class of Calculator", 
  test_case=TestCalculatorSub
)



import subprocess
import os
from pathlib import Path
from .utils import Color
import shutil
import yaml
import sys
color = Color()
cli_path = Path(os.path.dirname(__file__)).resolve()
project_path = Path(os.getcwd()).resolve()

class Initializer:
    def __init__(self):
        pass

    def create(self):
        user_input = self.get_user_input()
        # create global .logicbook dir
        self.create_config_dir(user_input)
        # create logics dir
        self.create_logics_dir(user_input)
        color.green("Initialization finished.")
        color.blue("Please run 'logicbook start' to start the server.")

    def get_user_input(self):
        user_input = {}
        color.green("Please enter your project name: [default is '']")
        user_input["project_name"] = str(input("Project Name: ") or "")
        color.green("Please enter your src path: [default is /]")
        color.blue("Tip: logics directory will be created under the entered src path.")
        path = str(input("Src Path: ") or "")
        if path.startswith("/"):
            path = path[1:]
        user_input["path"] = path
        return user_input

    def create_config_dir(self, user_input):
        color.green("Creating .logicbook dir")
        # create global .logicbook dir
        source_dir_path = cli_path.joinpath('../template/.logicbook').resolve()
        target_dir_path = project_path.joinpath('.logicbook').resolve()
        shutil.copytree(source_dir_path, target_dir_path)

        config = None
        with open(target_dir_path.joinpath('config.yaml'), 'r') as f:
            config = yaml.safe_load(f)
            config['project_name'] = user_input['project_name']
            config['path'] = user_input['path']
        with open(target_dir_path.joinpath("config.yaml"), 'w') as f:
            yaml.dump(config, f)

    def create_logics_dir(self, user_input):
        color.green("Creating logics dir")
        # create logics dir
        source_dir_path = cli_path.joinpath('../template/logics').resolve()
        target_dir_path = project_path.joinpath(user_input["path"]).joinpath('logics').resolve()
        shutil.copytree(source_dir_path, target_dir_path)



import unittest
from calculator import Calculator

class TestCalculatorAdd(unittest.TestCase):

    def setUp(self):
        self.calc = Calculator()

    def test_add(self):
        actual = self.calc.add(1, 2)
        self.assertEqual(3, actual)

class TestCalculatorSub(unittest.TestCase):

    def setUp(self):
        self.calc = Calculator()

    def test_sub(self):
        actual = self.calc.sub(2, 1)
        self.assertEqual(1, actual)

if __name__ == '__main__':
    unittest.main()


import os
import inspect

def test():
    print("test")

print(os.path.abspath(inspect.getfile(test)))

class Test:
    def __init__(self) -> None:
        pass

    def test1(self):
        print("test1")

    def test2(self):
        print("test2")

import unittest
def sum_logic(a, b):
    return a + b

class TestSum(unittest.TestCase):
    def setUp(self, a):
        self.a = a

    def test_sum(self):
        num = self.a
        actual = sum_logic(1, 2)
        self.assertEqual(3, actual)

def getLocalMethods(clss):
    import types
    result = [ ]
    print(clss.__dict__, types.FunctionType)
    for var in clss.__dict__:
        val = clss.__dict__[var]
        if isinstance(val, types.FunctionType) and var.startswith('test'):
            result.append(var)
    return sorted(result)
    
print(Test)
print(getLocalMethods(Test))
print(getLocalMethods(TestSum))
methods = getLocalMethods(TestSum)
for method in methods:
    print(method)
    print(TestSum.__dict__[method])
    print(TestSum.__dict__[method].__name__)

print(hasattr(getLocalMethods, '__self__'))

def foo(a, b, c=4, *arglist, **keywords): pass

class Test:
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def test1(self):
        return self.a + self.b

tes = Test(1, 2)
print(tes.__dict__, tes.test1())
args = tes.__dict__
tes2 = Test(**args)
print(tes2.__dict__, tes2.test1())
print(inspect.getfullargspec(tes.__init__))

print(tes.test1)
print(Test.test1)
#test_sum = TestSum()
#print(test_sum.test_sum())


from numpy import diff
from fastapi import FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from starlette.middleware.cors import CORSMiddleware
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
import os
import subprocess
import sys
import time
import threading
from pydantic import BaseModel
from manager import LogicManager
from logicbook import Logic  # This may be wrong
import importlib
from pathlib import Path
import inspect
from utils import Color
from config import get_config

core_dir = Path(os.path.dirname(__file__)).resolve()
project_dir = Path(os.getcwd()).resolve()
color = Color()

app = FastAPI(
    title="Logicbook",
)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,   # 追記により追加
    allow_methods=["*"],      # 追記により追加
    allow_headers=["*"]       # 追記により追加
)
app.mount("/static", StaticFiles(directory=core_dir.joinpath("../ui/build/static").resolve()), name="static")
templates = Jinja2Templates(directory=core_dir.joinpath("../ui/build").resolve())
# index page
@app.get("/")
async def serve_ui(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})
# logic page
@app.get("/logics/{name}")
async def serve_ui2(request: Request, name: str):
    return templates.TemplateResponse("index.html", {"request": request})

app.manager = LogicManager()
alive = True
@app.on_event("startup")
async def startup_event():
    # run watcher
    watcher_thread = threading.Thread(target=run_watcher)
    watcher_thread.start()
    # add logic
    config = get_config(project_dir.joinpath(".logicbook/config.yaml"))
    logics_dir = project_dir.joinpath(config.path).joinpath('logics').resolve()
    for file in logics_dir.glob('**/*_book.py'):
        sys.path.append(str(file.parent)) # TODO: remove pycache
        module_name = file.name.split('.')[0]
        module = importlib.import_module(module_name, file.parent)
        
        # add logic
        for name in dir(module):
            if isinstance(getattr(module, name), Logic):
                color.blue(f"Loading {name} at {file.parent}")
                logic = getattr(module, name)
                logic.add_book_logic_name(name)
                app.manager.add_logic(logic)

@app.on_event("shutdown")
def shutdown_event():
    color.green(f"Stopping Logicbook server")
    global alive
    alive = False

class ExecuteLogicModel(BaseModel):
    id: str
    args : dict

@app.post("/execute/logic")
async def execute_logic(data: ExecuteLogicModel):
    result = None
    for logic in app.manager.logics:
        if logic.id == data.id:
            logic.reload()
            result = logic.func(**data.args)
    return result

@app.post("/execute/example")
async def execute_example(data: ExecuteLogicModel):
    result = app.manager.execute_example(data.id, data.args)
    return result

class ExecuteTestModel(BaseModel):
    id: str

@app.post("/execute/test")
async def execute_test(data: ExecuteTestModel):
    result = app.manager.execute_test(data.id)
    return result
    '''for logic in app.manager.logics:
        for test in logic.tests:
            if test.id == data.id:
                test.run()
    return None'''

@app.get("/data")
async def get_data():
    data = [logic.json() for logic in app.manager.logics]
    return data

def run_watcher():
    target_dir = project_dir
    event_handler = FileChangeHandler()
    observer = Observer()
    observer.schedule(event_handler, target_dir, recursive=True)
    observer.start()
    try:
        global arive
        while alive:
            time.sleep(0.1)
    except KeyboardInterrupt:
        observer.stop()
    observer.stop()
    observer.join()


class Time:
    def __init__(self):
        self.time = time.time()
        self.lock = threading.Lock()

    def set(self, _time):
            self.lock.acquire() # 排他制御開始
            self.time = _time
            self.lock.release() # 排他制御解除

app.time = Time()
class FileChangeHandler(FileSystemEventHandler):

    def on_created(self, event):
        filepath = event.src_path
        filename = os.path.basename(filepath)
        #print('%s created' % filename)

    def on_modified(self, event):
        curren_time = time.time()
        diff_time = curren_time - app.time.time
        app.time.set(curren_time)
        if diff_time < 0.1: # for fix called twice bug
            filepath = event.src_path
            filename = os.path.basename(filepath)

            for logic in app.manager.logics:
                # update logic if logic file is changed
                if str(logic.func_path) == filepath or str(logic.book_path) == filepath or str(logic.readme_path) == filepath:
                    if logic.is_changed():
                        color.green("Update Logic: {} at {}".format(logic.func.__name__, filename))
                        logic.update()
                    continue
                # run test if test file is changed
                for test in logic.tests:
                    if str(test.path) == filepath:
                        if test.is_changed():
                            color.green("Update Test: {} at {}".format(test.func.__name__, filename))
                            test.run()
                            test.update()
                        continue

    def on_deleted(self, event):
        filepath = event.src_path
        filename = os.path.basename(filepath)
        #print('%s deleted' % filename)

    def on_moved(self, event):
        filepath = event.src_path
        filename = os.path.basename(filepath)
        #print('%s moved' % filename)



from join import join
from logicbook import Logic
from test_join import TestJoin

mylogic = Logic(
    name="Join",
    func=join,
    readme="join.md",
)

mylogic.add_example(
    name="Default", 
    func=join,
    args={
        "a": "Hello ",
        "b": "Logicbook!"
    }
)

mylogic.add_test(
  name="Test Class of Join", 
  test_case=TestJoin
)


import unittest
from join import join

class TestJoin(unittest.TestCase):

    def test_join(self):
        actual = join(1, 2)
        self.assertEqual(3, actual)

if __name__ == '__main__':
    unittest.main()


import click
from .initializer import Initializer
from .runner import Runner
from pathlib import Path
import os

cli_path = Path(os.path.dirname(__file__)).resolve()

@click.group()
def cmd():
    pass

@cmd.command()
def init():
    i = Initializer()
    i.create()

@cmd.command()
@click.option('--port', '-p', default=8000, help="port number")
def start(port):
    runner = Runner()
    runner.run_server(port)

def main():
    cmd()
    
if __name__ == '__main__':
    main()


from setuptools import setup, find_packages

with open('README.rst') as f:
    readme = f.read()

with open('requirements.txt') as f:
    all_reqs = f.read().split('\n')
install_requires = [x.strip() for x in all_reqs]

setup(
    name="logicbook",
    version="0.1.4",
    packages=["logicbook"],
    include_package_data=True,
    entry_points={
            "console_scripts": ['logicbook = logicbook.cli.root:main']
        },
    install_requires = install_requires,
    description='Logicbook is an open source tool for building python logics in isolation. It streamlines logic development, testing, and documentation.',
    long_description=readme,
    author='Rui Hirano',
    license='MIT',
    url='https://ruihirano.github.io/logicbook/',
)



from datetime import datetime
from typing import List
import subprocess
import uuid
from pathlib import Path
import os
import inspect
import linecache
import importlib

core_dir = Path(os.path.dirname(__file__)).resolve()
project_dir = Path(os.getcwd()).resolve()

class LogicManager:
    def __init__(self):
        self.logics: List[Logic] = []

    def add_logic(self, logic):
        self.logics.append(logic)

    def get_logic(self, name):
        for logic in self.logics:
            if logic.name == name:
                return logic
        return None

    def execute_example(self, id, args):
        for logic in self.logics:
            for example in logic.examples:
                if example.id == id:
                    return example.run(args)
        return None

    def execute_test(self, id):
        for logic in self.logics:
            for test in logic.tests:
                if test.id == id:
                    return test.run()
        return None

def get_test_names(cls):
    import types
    result = [ ]
    for var in cls.__dict__:
        val = cls.__dict__[var]
        if isinstance(val, types.FunctionType) and var.startswith('test'):
            result.append(var)
    return sorted(result)

class Logic:
    def __init__(self, name, func, readme=None) -> None:
        self.id = str(uuid.uuid4())
        self.category, self.name= self.get_name_category(name)
        self.book_path = Path(os.path.abspath((inspect.stack()[1])[1]))
        self.book_filename = str(self.book_path.name).split('.')[0]
        self.book_module = importlib.import_module(self.book_filename, self.book_path.parent)
        self.book_code = inspect.getsource(self.book_module)
        #self.book_func = getattr(self.func_module, name)
        self.func_path = Path(os.path.abspath(inspect.getfile(func)))
        self.func_filename = str(self.func_path.name).split('.')[0]
        self.func_module = importlib.import_module(self.func_filename, self.func_path.parent)
        self.readme_path = self.book_path.parent.joinpath(readme).resolve() if readme != None else None
        self.func=func
        self.readme = self.get_markdown(self.readme_path) if self.readme_path != None else None
        self.tests=[]
        self.examples = []
        self.code = inspect.getsource(func)
        self.cls, self.cls_args = self.check_cls(func)

    def get_name_category(self, name):
        split_name = name.split('/')
        if len(split_name) == 1:
            return None, name
        else:
            return split_name[0], split_name[1]

    def add_book_logic_name(self, name):
        self.book_logic_name = name

    def check_cls(self, func):
        exist = False
        # check if func or not
        for name in dir(self.func_module):
            if name == func.__name__:
                exist = True
                return None, None
        # check if func in class or not
        for name in dir(self.func_module):
            if hasattr(func, '__self__') and hasattr(func.__self__, '__class__') and name == func.__self__.__class__.__name__:
                exist = True
                cls = func.__self__.__class__
                args_names = inspect.getargspec(cls.__init__).args[1:]
                self_dict = {k.strip('_'): v for k, v in func.__self__.__dict__.items()}
                args = {k: v for k, v in self_dict.items() if k in args_names}
                return cls, args
        if not exist:
            raise Exception(f"{func.__name__} is invalid function")

    def reload(self):
        self.func_module = importlib.reload(self.func_module)
        self.book_module = importlib.reload(self.book_module)
        for name in dir(self.func_module):
            if self.cls:
                if name == self.cls.__name__:
                    clas = getattr(self.func_module, name)(**self.cls_args)
                    self.func = getattr(clas, self.func.__name__)
            else:
                if name == self.func.__name__:
                    self.func = getattr(self.func_module, name)

    def is_changed(self):
        self.reload()
        func_str = inspect.getsource(self.func)
        is_changed_func = self.code != func_str
        is_changed_book = self.book_code != inspect.getsource(self.book_module) # TODO: check book
        is_changed_readme = self.readme != self.get_markdown(self.readme_path) if self.readme_path != None else None
        return is_changed_func or is_changed_readme or is_changed_book

    def update(self):
        self.readme = self.get_markdown(self.readme_path) if self.readme_path != None else None
        self.code = inspect.getsource(self.func)
        self.book_code = inspect.getsource(self.book_module)
        for test in self.tests:
            test.run()

    def get_markdown(self, path):
        md = None
        with open(path) as f:
            md = f.read()
        return md

    def add_example(self, name, func, args):
        self.examples.append(Example(name, func, args))

    def add_test(self, name, test_case):
        testnames = get_test_names(test_case)
        for name in testnames:
            func = test_case.__dict__[name]
            test = Test(name, test_case, func)
            test.run()
            self.tests.append(test)

    def json(self):
        return {
            "id": self.id,
            "category": self.category,
            "name": self.name,
            "book_path": str(self.book_path),
            "func_path": str(self.func_path),
            "readme_path": str(self.readme_path),
            "readme": self.readme,
            "tests": [test.json() for test in self.tests],
            "examples": [ex.json() for ex in self.examples],
            "code": self.code,
        }

class Example:
    def __init__(self, name, func, args):
        self.id = str(uuid.uuid4())
        self.name=name
        self.func=func
        self.args=args
        self.output=self.run(args)

    def json(self):
        return {
            "id": self.id,
            "name": self.name,
            "args": self.args,
            "output": self.output,
        }

    def run(self, args):
        try:
            self.output = str(self.func(**args))
        except Exception as e:
            self.output = str(e)
        return self.output

class Test:
    def __init__(self, name, test_case, func):
        self.id = str(uuid.uuid4())
        self.name=name
        self.test_case=test_case
        self.func=func
        self.func_path = Path(os.path.abspath(inspect.getfile(func)))
        self.func_filename = str(self.func_path.name).split('.')[0]
        self.func_module = importlib.import_module(self.func_filename, self.func_path.parent)
        self.path = os.path.abspath(inspect.getfile(func))
        self.status="unknown"
        self.code = inspect.getsource(func)
        self.result = None
        self.latest_run_time = None

    def json(self):
        return {
            "id": self.id,
            "name": self.name,
            "test_case_name": self.test_case.__name__,
            "path": self.path,
            "func_path": str(self.func_path),
            "code": self.code,
            "status": self.status,
            "result": self.result,
            "latest_run_time": self.latest_run_time.isoformat(),
        }

    def reload(self):
        self.func_module = importlib.reload(self.func_module)
        for name in dir(self.func_module):
            if name == self.test_case.__name__:
                clas = getattr(self.func_module, name)()
                self.func = getattr(clas, self.func.__name__)

    def is_changed(self):
        self.reload()
        return self.code != inspect.getsource(self.func)

    def update(self):
        self.code = inspect.getsource(self.func)

    def run(self):
        command = ["python3", self.path, "{}.{}".format(self.test_case.__name__, self.func.__name__)]
        proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout_data, stderr_data = proc.communicate()
        if proc.returncode == 0:
            self.status = "success"
        else:
            self.status = "failure"
        self.result = stderr_data.decode('utf-8')
        self.latest_run_time = datetime.now()


import os
import sys
sys.path.append(os.path.join(os.path.dirname(__file__), '../core'))

from pathlib import Path
import json
from .utils import Color
from ..core.server import app
import uvicorn
import webbrowser

color = Color()
project_path = Path(os.getcwd()).resolve()
ui_build_dir = Path(os.path.dirname(__file__)).joinpath("../ui/build").resolve()
class Runner:
    def __init__(self):
        pass

    def check_exist_config(self):
        # check .logicbook and config in root
        if not project_path.joinpath('.logicbook').exists():
            color.red("Error: .logicbook dir not found")
            sys.exit(1)
        if not project_path.joinpath('.logicbook/config.yaml').exists():
            color.red("Error: config.yaml not found in .logicbook dir")
            sys.exit(1)

    def run_server(self, port):
        self.check_exist_config()
        color.green("Starting the logicbook server at http://localhost:{}".format(port))
        webbrowser.open("http://localhost:{}".format(port))
        uvicorn.run(app, host="0.0.0.0", port=port, log_level="error")